/*
 * Copyright (c) 2007 Luigi Rizzo - Universita` di Pisa
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Id: usb.h,v 1.41 2008/04/08 18:56:24 luigi Exp $
 * Compat file for linux/usb.h
 *
 * Here we implement the data structures needed to emulate the linux
 * kernel API on FreeBSD. The main part of this file is documentation
 * on the different structures seen by usb devices in both kernels.
 * Then we have the definition of various linux data structures,
 * trimmed removing all parts not used by drivers.
 */

#ifndef _LINUX_USB_H
#define _LINUX_USB_H

#include <sys/module.h>		/* module_t */

#include <linux/version.h>	/* KERNEL_VERSION */
#include <linux/kref.h>             /* for struct kref */

#include <asm/types.h>
#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>

struct usb_device;
struct usb_interface;

#if 0 /* comments */

A bit of information on what (i think) are the usb architectures
on Linux and FreeBSD.

Each usb device has one or more interfaces. Each interface
can have up to 8 endpoints. An endpoint can have a few
alternate configurations.

===================== Linux Data Structures ======================

linux/device.h:
	struct device {
		...
		struct device   * parent;
		struct device_driver *driver;
		void            *driver_data;
		...
	}

	struct usb_host_interface {
		struct usb_interface_descriptor desc;
		/* array of desc.bNumEndpoint endpoints associated with this
		 * interface setting.  these will be in no particular order.
		 * XXX we need to allocate this.
		 */
		struct usb_host_endpoint *endpoint;
		...
	};

	struct usb_interface {
		...
		struct usb_host_interface *cur_altsetting; /* current setting */
		struct device dev;
		...
	}

	struct usb_device {
		...
		struct device dev;
		...
	}

There is one descriptor per interface (intf), linked to the parent
usb_device (udev) through the 'struct device' in each of them, as below:

	struct usb_interface *intf;	// an interface
	struct usb_device *udev;	// the parent device

	intf->dev.parent == &udev->dev;

We reach the container of the 'struct device' through the macro container_of().
At the beginning of the probe routine we have:

	struct usb_interface *intf;	// an interface
	struct usb_device *dev = interface_to_usbdev(intf);
		i.e. container of intf->dev.parent
		in FreeBSD, intf->device

	interface = &intf->cur_altsetting->desc;
		in FreeBSD, interface = intf->idesc

In order to emulate these data structures, we need to provide the following
data structures, appropriately linked. In our emulation layer, these
are allocated as part of the softc, within the structure

    struct linux_usb_descriptors {
	...
	struct file
	struct usb_interface	for the interface in use
		must be persistent
	struct usb_device	for the parent device ?
		must be persistent
	struct usb_host_interface	for the (current ?altsetting)
	struct usb_host_endpoint[N]	for the endpoints.
    }

which is defined in ldev_stub.h (XXX should be elsewhere).
-- Descriptor linkage: ---

We refer here to linux videodev devices:

	softc...{  <-----------------------+
	.  vdev .-->	video_device....{  |
	.	.	. priv ---------.--+
	.	.	. fops ---------.-->	file_operations {
	.  dev -->-+	. dev  -->--+	.	    open, close, ioctl...
	}........  |	}...........|....	}
	           |		    |
		   +----------------+
				    |
	linux_usb_descriptors {     |
				    |
				    |
		usb_interface {	<--------- intf to linux probe routine
				    |
		    device {        |
			parent -->--+
		    }		    |
				    |
		}		    |
				    |
		usb_device {	    |
				    |
		    device {  <-----+
		    }

		}
	}
		

===================== FreeBSD Data Structures ======================

    (note, device_t and device_ptr_t are actually the same thing)

    Most descriptors are 'struc usbd_foo' and have a matching
    typedef-d pointer typedef struct usbd_foo * usbd_foo_handle
    defined in usbdi.h

sys/dev/usb/usbdivar.h:

    struct usbd_device {	// ptd by usbd_device_handle
	...
	struct usbd_pipe       *default_pipe; // pipe 0
	u_int8_t                config;        // current configuration # */
	...
	struct usbd_endpoint    def_ep;        /* for pipe 0 */
	usb_endpoint_descriptor_t def_ep_desc; /* for pipe 0 */
	struct usbd_interface  *ifaces;        /* array of all interfaces */
	usb_device_descriptor_t ddesc;         /* device descriptor */
	usb_config_descriptor_t *cdesc;        /* full config descr */
	device_ptr_t           *subdevs;       /* sub-devices, 0 terminated */
	uint8_t                *ifacenums;     /* sub-device interfacenumbers */
    }

    struct usbd_interface {
	struct usbd_device     *device;
	usb_interface_descriptor_t *idesc;
	int                     index;
	int                     altindex;
	struct usbd_endpoint   *endpoints;
	void                   *priv;
	LIST_HEAD(, usbd_pipe)  pipes;
    };

    struct usbd_pipe {
	struct usbd_interface  *iface;
	struct usbd_device     *device;
	struct usbd_endpoint   *endpoint;
	int                     refcnt;
	char                    running;
	char                    aborting;
	SIMPLEQ_HEAD(, usbd_xfer) queue;
	LIST_ENTRY(usbd_pipe)   next;

	usbd_xfer_handle        intrxfer; /* used for repeating requests */
	char                    repeat;
	int                     interval;

	/* Filled by HC driver. */
	struct usbd_pipe_methods *methods;
    };

sys/dev/usb/usbdi.h:
    /* Attach data */
    struct usb_attach_arg {
	int                     port;
	int                     configno;
	int                     ifaceno;
	int                     vendor;
	int                     product;
	int                     release;
	int                     matchlvl;
	usbd_device_handle      device; /* current device */
	usbd_interface_handle   iface; /* current interface */
	int                     usegeneric;
	usbd_interface_handle  *ifaces; /* all interfaces */
	int                     nifaces; /* number of interfaces */
    };

The match/attach routines are passed a device_t dev argument,
which can access the softc and attach arg with

	struct ldev_softc *sc = device_get_softc(dev);
	struct usb_attach_arg *uaa = device_get_ivars(dev);

and from the latter you can get the interface, the parent device and so on.
It is probably a good idea to store the device/interface in the
softc during the attach process:

        sc->udev = uaa->device;
	sc->sc_iface = uaa->iface;
        sc->sc_dev = dev;

In the other routines, you get struct cdev *dev [the naming is confusing]
as first argument; from here you can do

	int unit = LDEVUNIT(dev);
        struct ldev_softc *sc = devclass_get_softc(ldev_devclass, unit);

and from the softc you can access the linux emulation fields.

#endif /* COMMENTS */

#define interface_to_usbdev(intf) \
        container_of(intf->dev.parent, struct usb_device, dev)


/* in mod_devicetable.h */
/**
 * Device table entry for "new style" table-driven USB drivers.
 * User mode code can read these tables to choose which modules to load.
 * Declare the table as a MODULE_DEVICE_TABLE.
 *
 * A probe() parameter will point to a matching entry from this table.
 * Use the driver_info field for each match to hold information tied
 * to that match:  device quirks, etc.
 *
 * Terminate the driver's table with an all-zeroes entry.
 * Use the flag values to control which fields are compared.
 */

/**
 * struct usb_device_id - identifies USB devices for probing and hotplugging
 * @match_flags: Bit mask controlling of the other fields are used to match
 *      against new devices.  Any field except for driver_info may be used,
 *      although some only make sense in conjunction with other fields.
 *      This is usually set by a USB_DEVICE_*() macro, which sets all
 *      other fields in this structure except for driver_info.
 * @idVendor: USB vendor ID for a device; numbers are assigned
 *      by the USB forum to its members.
 * @idProduct: Vendor-assigned product ID.
 * @bcdDevice_lo: Low end of range of vendor-assigned product version numbers.
 *      This is also used to identify individual product versions, for
 *      a range consisting of a single device.
 * @bcdDevice_hi: High end of version number range.  The range of product
 *      versions is inclusive.
 * @bDeviceClass: Class of device; numbers are assigned
 *      by the USB forum.  Products may choose to implement classes,
 *      or be vendor-specific.  Device classes specify behavior of all
 *      the interfaces on a devices.
 * @bDeviceSubClass: Subclass of device; associated with bDeviceClass.
 * @bDeviceProtocol: Protocol of device; associated with bDeviceClass.
 * @bInterfaceClass: Class of interface; numbers are assigned
 *      by the USB forum.  Products may choose to implement classes,
 *      or be vendor-specific.  Interface classes specify behavior only
 *      of a given interface; other interfaces may support other classes.
 * @bInterfaceSubClass: Subclass of interface; associated with bInterfaceClass.
 * @bInterfaceProtocol: Protocol of interface; associated with bInterfaceClass.
 * @driver_info: Holds information used by the driver.  Usually it holds
 *      a pointer to a descriptor understood by the driver, or perhaps
 *      device flags.
 *
 * In most cases, drivers will create a table of device IDs by using
 * USB_DEVICE(), or similar macros designed for that purpose.
 * They will then export it to userspace using MODULE_DEVICE_TABLE(),
 * and provide it to the USB core through their usb_driver structure.
 *
 * See the usb_match_id() function for information about how matches are
 * performed.  Briefly, you will normally use one of several macros to help
 * construct these entries.  Each entry you provide will either identify
 * one or more specific products, or will identify a class of products
 * which have agreed to behave the same.  You should put the more specific
 * matches towards the beginning of your table, so that driver_info can
 * record quirks of specific products.
 */
struct usb_device_id {
        /* which fields to match against? */
        __u16           match_flags;

        /* Used for product specific matches; range is inclusive */
        __u16           idVendor;
        __u16           idProduct;
        __u16           bcdDevice_lo;
        __u16           bcdDevice_hi;

        /* Used for device class matches */
        __u8            bDeviceClass;
        __u8            bDeviceSubClass;
        __u8            bDeviceProtocol;

        /* Used for interface class matches */
        __u8            bInterfaceClass;
        __u8            bInterfaceSubClass;
        __u8            bInterfaceProtocol;

        /* not matched against */
        kernel_ulong_t  driver_info;
};

#define USB_DEVICE_ID_MATCH_VENDOR              0x0001
#define USB_DEVICE_ID_MATCH_PRODUCT             0x0002
#define USB_DEVICE_ID_MATCH_DEV_LO              0x0004
#define USB_DEVICE_ID_MATCH_DEV_HI              0x0008
#define USB_DEVICE_ID_MATCH_DEV_CLASS           0x0010
#define USB_DEVICE_ID_MATCH_DEV_SUBCLASS        0x0020
#define USB_DEVICE_ID_MATCH_DEV_PROTOCOL        0x0040
#define USB_DEVICE_ID_MATCH_INT_CLASS           0x0080
#define USB_DEVICE_ID_MATCH_INT_SUBCLASS        0x0100
#define USB_DEVICE_ID_MATCH_INT_PROTOCOL        0x0200

#define USB_DEVICE_ID_MATCH_DEVICE \
	(USB_DEVICE_ID_MATCH_VENDOR | USB_DEVICE_ID_MATCH_PRODUCT)

#define USB_DEVICE_ID_MATCH_INT_INFO \
                (USB_DEVICE_ID_MATCH_INT_CLASS | \
                USB_DEVICE_ID_MATCH_INT_SUBCLASS | \
                USB_DEVICE_ID_MATCH_INT_PROTOCOL)

#define USB_DEVICE(vend,prod) \
	.match_flags = USB_DEVICE_ID_MATCH_DEVICE, .idVendor = (vend), \
	.idProduct = (prod)

/**
 * USB_INTERFACE_INFO - macro used to describe a class of usb interfaces
 * @cl: bInterfaceClass value
 * @sc: bInterfaceSubClass value
 * @pr: bInterfaceProtocol value
 *  
 * This macro is used to create a struct usb_device_id that matches a
 * specific class of interfaces.
 */
#define USB_INTERFACE_INFO(cl,sc,pr) \
        .match_flags = USB_DEVICE_ID_MATCH_INT_INFO, .bInterfaceClass = (cl), \
        .bInterfaceSubClass = (sc), .bInterfaceProtocol = (pr)

#if 0
struct device_driver {	// dummy, to be completed
};

struct usb_dynids {
        spinlock_t lock;
        struct list_head list;
};
#endif

struct usb_driver {
        const char *name;

        int (*probe) (struct usb_interface *intf,
                      const struct usb_device_id *id);

        void (*disconnect) (struct usb_interface *intf);

#if 0	/* not currently used */
        int (*ioctl) (struct usb_interface *intf, unsigned int code,
                        void *buf);

        int (*suspend) (struct usb_interface *intf, pm_message_t message);
        int (*resume) (struct usb_interface *intf);
#endif	/* not currently used */

        const struct usb_device_id *id_table;
#if 0	/* not currently used */
        struct usb_dynids dynids;
        struct device_driver driver;
        unsigned int no_dynamic_id:1;
#endif	/* not currently used */
};

/*
 * When a device registers (usually from the .init function,
 * which in our emulation is called from modinit_handler()),
 * we store the driver info in the structure containing the
 * other device information.
 * On deregister, we don't do anything - in principle we could
 * check that the device is registered and remove the entry.
 */
static inline int usb_register(struct usb_driver *driver)
{
	if (MODINFO_NAME(info).usb_driver_desc != NULL)
		return -EINVAL;	/* already registered */
	MODINFO_NAME(info).usb_driver_desc = driver;
	return 0;
}

void usb_deregister(struct usb_driver *);

/*
 * All standard descriptor start with bLength and bDescriptorType
 * followed by descriptor-specific stuff. For feature 'foo':
 *
 * on linux there is a struct usb_foo_descriptor, and a macro
 * USB_DT_FOO_SIZE indicating the size.
 *
 * on FreeBSD there is a typedef usb_foo_descriptor_t, and a macro
 * USB_FOO_DESCRIPTOR_SIZE indicating the size.
 *
 * The structure is the same, so we can cast or copy between them.
 * Beware though that on FreeBSD __le16 becomes uWord, which is
 * an array of bytes (little endian) that needs to be read with
 * UGETW and writtend with USETW.
 *
 *	Linux			FreeBSD
 *	device			device
 *	config			config
 *	string			string [127max ]
 *	interface		interface
 *	endpoint		endpoint
 *	qualifier		device_qualifier (non-std name)
 *	otg			otg
 *	interface_assoc
 *	security
 *	key
 *	encryption
 *	bos
 *	cap
 *	wireless_cap
 *	wireless_ep
 */

/*
 * Descriptor types ... USB 2.0 spec table 9.5
 */
#define USB_DT_DEVICE                   0x01
#define USB_DT_CONFIG                   0x02
#define USB_DT_STRING                   0x03
#define USB_DT_INTERFACE                0x04
#define USB_DT_ENDPOINT                 0x05
#define USB_DT_DEVICE_QUALIFIER         0x06
#define USB_DT_OTHER_SPEED_CONFIG       0x07
#define USB_DT_INTERFACE_POWER          0x08
/* these are from a minor usb 2.0 revision (ECN) */
#define USB_DT_OTG                      0x09
#define USB_DT_DEBUG                    0x0a
#define USB_DT_INTERFACE_ASSOCIATION    0x0b
/* these are from the Wireless USB spec */
#define USB_DT_SECURITY                 0x0c
#define USB_DT_KEY                      0x0d
#define USB_DT_ENCRYPTION_TYPE          0x0e
#define USB_DT_BOS                      0x0f
#define USB_DT_DEVICE_CAPABILITY        0x10
#define USB_DT_WIRELESS_ENDPOINT_COMP   0x11
#define USB_DT_WIRE_ADAPTER             0x21
#define USB_DT_RPIPE                    0x22

/* conventional codes for class-specific descriptors */
#define USB_DT_CS_DEVICE                0x21
#define USB_DT_CS_CONFIG                0x22
#define USB_DT_CS_STRING                0x23
#define USB_DT_CS_INTERFACE             0x24
#define USB_DT_CS_ENDPOINT              0x25
 

struct usb_device_descriptor {
        __u8  bLength;
        __u8  bDescriptorType;

        __le16 bcdUSB;
        __u8  bDeviceClass;
        __u8  bDeviceSubClass;
        __u8  bDeviceProtocol;
        __u8  bMaxPacketSize0;
        __le16 idVendor;
        __le16 idProduct;
        __le16 bcdDevice;
        __u8  iManufacturer;
        __u8  iProduct;
        __u8  iSerialNumber;
        __u8  bNumConfigurations;
} __attribute__ ((packed));

#define USB_DT_DEVICE_SIZE              18  

/*
 * Device and/or Interface Class codes
 * as found in bDeviceClass or bInterfaceClass
 * and defined by www.usb.org documents
 */
#define USB_CLASS_PER_INTERFACE         0       /* for DeviceClass */
#define USB_CLASS_AUDIO                 1
#define USB_CLASS_COMM                  2
#define USB_CLASS_HID                   3
#define USB_CLASS_PHYSICAL              5
#define USB_CLASS_STILL_IMAGE           6
#define USB_CLASS_PRINTER               7
#define USB_CLASS_MASS_STORAGE          8
#define USB_CLASS_HUB                   9
#define USB_CLASS_CDC_DATA              0x0a
#define USB_CLASS_CSCID                 0x0b    /* chip+ smart card */
#define USB_CLASS_CONTENT_SEC           0x0d    /* content security */
#define USB_CLASS_VIDEO                 0x0e
#define USB_CLASS_WIRELESS_CONTROLLER   0xe0
#define USB_CLASS_APP_SPEC              0xfe
#define USB_CLASS_VENDOR_SPEC           0xff


/* USB_DT_CONFIG: Configuration descriptor information.
 *
 * USB_DT_OTHER_SPEED_CONFIG is the same descriptor, except that the
 * descriptor type is different.  Highspeed-capable devices can look
 * different depending on what speed they're currently running.  Only   
 * devices with a USB_DT_DEVICE_QUALIFIER have any OTHER_SPEED_CONFIG
 * descriptors.
 */
struct usb_config_descriptor {
        __u8  bLength;
        __u8  bDescriptorType;

        __le16 wTotalLength;
        __u8  bNumInterfaces;
        __u8  bConfigurationValue;
        __u8  iConfiguration;
        __u8  bmAttributes;
        __u8  bMaxPower;
} __attribute__ ((packed));

#define USB_DT_CONFIG_SIZE              9

/* USB_DT_STRING: String descriptor */
struct usb_string_descriptor {
        __u8  bLength;
        __u8  bDescriptorType;
        
        __le16 wData[1];                /* UTF-16LE encoded */
} __attribute__ ((packed));

/* note that "string" zero is special, it holds language codes that
 * the device supports, not Unicode characters.
 */

/* USB_DT_INTERFACE: Interface descriptor */
struct usb_interface_descriptor {
        __u8  bLength;
        __u8  bDescriptorType;

        __u8  bInterfaceNumber;
        __u8  bAlternateSetting;
        __u8  bNumEndpoints;
        __u8  bInterfaceClass;
        __u8  bInterfaceSubClass;
        __u8  bInterfaceProtocol;
        __u8  iInterface;
} __attribute__ ((packed));

#define USB_DT_INTERFACE_SIZE           9

/* USB_DT_ENDPOINT: Endpoint descriptor */
struct usb_endpoint_descriptor {
        __u8  bLength;
        __u8  bDescriptorType;

        __u8  bEndpointAddress;
        __u8  bmAttributes;
        __le16 wMaxPacketSize;	/* uWord on FreeBSD */
        __u8  bInterval;

        /* NOTE:  these two are _only_ in audio endpoints. */
        /* use USB_DT_ENDPOINT*_SIZE in bLength, not sizeof. */
        __u8  bRefresh;
        __u8  bSynchAddress;
} __attribute__ ((packed));

/* size in bytes of the above structure, without or with audio extension */
#define USB_DT_ENDPOINT_SIZE            7
#define USB_DT_ENDPOINT_AUDIO_SIZE      9       /* Audio extension */

int usb_endpoint_dir_in(const struct usb_endpoint_descriptor *epd);
int usb_endpoint_xfer_int(const struct usb_endpoint_descriptor *epd);
int usb_endpoint_is_int_in(const struct usb_endpoint_descriptor *epd);

/*
 * Endpoints
 */
#define USB_ENDPOINT_NUMBER_MASK        0x0f    /* in bEndpointAddress */
#define USB_ENDPOINT_DIR_MASK           0x80

#define USB_ENDPOINT_XFERTYPE_MASK      0x03    /* in bmAttributes */
#define USB_ENDPOINT_XFER_CONTROL       0
#define USB_ENDPOINT_XFER_ISOC          1
#define USB_ENDPOINT_XFER_BULK          2
#define USB_ENDPOINT_XFER_INT           3
#define USB_ENDPOINT_MAX_ADJUSTABLE     0x80

struct usb_host_interface {
        struct usb_interface_descriptor desc;
        /* array of desc.bNumEndpoint endpoints associated with this
         * interface setting.  these will be in no particular order.
         */
        struct usb_host_endpoint *endpoint;

#if 0
        char *string;           /* iInterface string, if present */
#endif
	/* XXX need to fill them for uvc */
        unsigned char *extra;   /* Extra descriptors */
        int extralen;
};

int usb_control_msg(struct usb_device *dev, unsigned int pipe,
	__u8 request, __u8 requesttype, __u16 value, __u16 index,
	void *data, __u16 size, int timeout);

/* CONTROL REQUEST SUPPORT */

/*
 * USB directions
 *
 * This bit flag is used in endpoint descriptors' bEndpointAddress field.
 * It's also one of three fields in control requests bRequestType:
 *      dtt...rr
 * on the right we indicate the equivalent FreeBSD macro
 */
#define USB_DIR_OUT	0	/* to device, UT_WRITE	*/
#define USB_DIR_IN	0x80	/* to host, UT_READ	*/

/*
 * USB types, the second of three bRequestType fields
 */
#define USB_TYPE_MASK		(0x03 << 5)	// 0x60
#define USB_TYPE_STANDARD	(0x00 << 5)	// 0x00, UT_STANDARD
#define USB_TYPE_CLASS		(0x01 << 5)	// 0x20, UT_CLASS
#define USB_TYPE_VENDOR		(0x02 << 5)	// 0x40, UT_VENDOR
#define USB_TYPE_RESERVED	(0x03 << 5)	// 0x60

/*
 * USB recipients, the third of three bRequestType fields
 */
#define USB_RECIP_MASK                  0x1f
#define USB_RECIP_DEVICE                0x00	// UT_DEVICE
#define USB_RECIP_INTERFACE             0x01	// UT_INTERFACE
#define USB_RECIP_ENDPOINT              0x02	// UT_ENDPOINT
#define USB_RECIP_OTHER                 0x03	// UT_OTHER

/*
 * On FreeBSD there are also macros that combine the
 * three fields above e.g UT_READ_CLASS_DEVICE...
 */

/*
 * Standard requests, for the bRequest field of a SETUP packet.
 *
 * These are qualified by the bRequestType field, so that for example
 * TYPE_CLASS or TYPE_VENDOR specific feature flags could be retrieved
 * by a GET_STATUS request.
 * On FreeBSD, these macro replace USB_REQ_ with UR_
 */
#define USB_REQ_GET_STATUS              0x00	// UR_GET_STATUS
#define USB_REQ_CLEAR_FEATURE           0x01	// ...
#define USB_REQ_SET_FEATURE             0x03
#define USB_REQ_SET_ADDRESS             0x05
#define USB_REQ_GET_DESCRIPTOR          0x06
/*
 * On FreeBSD there is a set of UDESC_* macros that indicate
 * descriptor types in the GET/SET_DESCRIPTOR commands.
 */
#define USB_REQ_SET_DESCRIPTOR          0x07
#define USB_REQ_GET_CONFIGURATION       0x08
#define USB_REQ_SET_CONFIGURATION       0x09
#define USB_REQ_GET_INTERFACE           0x0A
#define USB_REQ_SET_INTERFACE           0x0B
#define USB_REQ_SYNCH_FRAME             0x0C

/* These are not present in the FreeBSD version */
#define USB_REQ_SET_ENCRYPTION          0x0D    /* Wireless USB */
#define USB_REQ_GET_ENCRYPTION          0x0E
#define USB_REQ_SET_HANDSHAKE           0x0F
#define USB_REQ_GET_HANDSHAKE           0x10
#define USB_REQ_SET_CONNECTION          0x11
#define USB_REQ_SET_SECURITY_DATA       0x12
#define USB_REQ_GET_SECURITY_DATA       0x13
#define USB_REQ_SET_WUSB_DATA           0x14
#define USB_REQ_LOOPBACK_DATA_WRITE     0x15
#define USB_REQ_LOOPBACK_DATA_READ      0x16
#define USB_REQ_SET_INTERFACE_DS        0x17

/*
 * USB feature flags are written using USB_REQ_{CLEAR,SET}_FEATURE, and
 * are read as a bit array returned by USB_REQ_GET_STATUS.  (So there
 * are at most sixteen features of each type.)
 * On FreeBSD, 
 */
#define USB_DEVICE_SELF_POWERED         0       /* (read only) */
#define USB_DEVICE_REMOTE_WAKEUP        1       /* dev may initiate wakeup */
#define USB_DEVICE_TEST_MODE            2       /* (wired high speed only) */
#define USB_DEVICE_BATTERY              2       /* (wireless) */
#define USB_DEVICE_B_HNP_ENABLE         3       /* (otg) dev may initiate HNP */
#define USB_DEVICE_WUSB_DEVICE          3       /* (wireless)*/
#define USB_DEVICE_A_HNP_SUPPORT        4       /* (otg) RH port supports HNP */
#define USB_DEVICE_A_ALT_HNP_SUPPORT    5       /* (otg) other RH port does */
#define USB_DEVICE_DEBUG_MODE           6       /* (special devices only) */

#define USB_ENDPOINT_HALT		0       /* IN/OUT will STALL */

// for more see http://www.gelato.unsw.edu.au/lxr/source/include/linux/usb_ch9.h#L35

enum usb_device_speed {
        USB_SPEED_UNKNOWN = 0,                  /* enumerating */
#if 0	/* these are already #defined on FreeBSD in sys/dev/usb/usb.h */
        USB_SPEED_LOW, USB_SPEED_FULL,          /* usb 1.1 */ 
        USB_SPEED_HIGH,                         /* usb 2.0 */
#endif
        USB_SPEED_VARIABLE = 4,			/* wireless (usb 2.5) */
};

enum usb_device_state {
        /* NOTATTACHED isn't in the USB spec, and this state acts
         * the same as ATTACHED ... but it's clearer this way.
         */
        USB_STATE_NOTATTACHED = 0,

        /* chapter 9 and authentication (wireless) device states */
        USB_STATE_ATTACHED,
        USB_STATE_POWERED,                      /* wired */
        USB_STATE_UNAUTHENTICATED,              /* auth */
        USB_STATE_RECONNECTING,                 /* auth */
        USB_STATE_DEFAULT,                      /* limited function */
        USB_STATE_ADDRESS,   
        USB_STATE_CONFIGURED,                   /* most functions */

        USB_STATE_SUSPENDED

        /* NOTE:  there are actually four different SUSPENDED
         * states, returning to POWERED, DEFAULT, ADDRESS, or
         * CONFIGURED respectively when SOF tokens flow again.
         */
};

#define __create_pipe(dev, endpoint)	(endpoint)
#define PIPE_ISOCHRONOUS                0
#define PIPE_INTERRUPT                  1
#define PIPE_CONTROL                    2
#define PIPE_BULK                       3

/*
 * On linux, pipes have the type and direction encoded in.
 * On FreeBSD, they don't.
 */
#define usb_sndctrlpipe(dev,endpoint)   \
	((PIPE_CONTROL << 30) | __create_pipe(dev,endpoint))

#define usb_rcvctrlpipe(dev,endpoint)   \
	((PIPE_CONTROL << 30) | __create_pipe(dev,endpoint) | USB_DIR_IN)

#define usb_sndisocpipe(dev,endpoint)   \
	((PIPE_ISOCHRONOUS << 30) | __create_pipe(dev,endpoint))

#define usb_rcvisocpipe(dev,endpoint)   \
	((PIPE_ISOCHRONOUS << 30) | __create_pipe(dev,endpoint) | USB_DIR_IN)

#define usb_sndbulkpipe(dev,endpoint)   \
	((PIPE_BULK << 30) | __create_pipe(dev,endpoint))

#define usb_rcvbulkpipe(dev,endpoint)   \
	((PIPE_BULK << 30) | __create_pipe(dev,endpoint) | USB_DIR_IN)

#define usb_sndintpipe(dev,endpoint)    \
	((PIPE_INTERRUPT << 30) | __create_pipe(dev,endpoint))

#define usb_rcvintpipe(dev,endpoint)    \
	((PIPE_INTERRUPT << 30) | __create_pipe(dev,endpoint) | USB_DIR_IN)

/* used these for multi-interface device registration */
int usb_driver_claim_interface(struct usb_driver *driver,
                        struct usb_interface *iface, void* priv);
 
/**
 * usb_interface_claimed - returns true iff an interface is claimed
 * @iface: the interface being checked
 *
 * Returns true (nonzero) iff the interface is claimed, else false (zero).
 * Callers must own the driver model's usb bus readlock.  So driver
 * probe() entries don't need extra locking, but other call contexts
 * may need to explicitly claim that lock.
 *
 */
int usb_interface_claimed(struct usb_interface *iface);


struct usb_interface *usb_ifnum_to_if(struct usb_device *dev, unsigned ifnum);
struct usb_host_interface *usb_altnum_to_altsetting(
                const struct usb_interface *intf, unsigned int altnum);
   
struct usb_host_endpoint {
        struct usb_endpoint_descriptor  desc;
#if 0
        struct list_head                urb_list;
        void                            *hcpriv;
        struct kobject                  *kobj;  /* For sysfs info */

        unsigned char *extra;   /* Extra descriptors */
        int extralen;
#endif
};

/* This is arbitrary.
 * From USB 2.0 spec Table 11-13, offset 7, a hub can
 * have up to 255 ports. The most yet reported is 10.
 */
#define USB_MAXCHILDREN         (16)

// struct usb_tt;

struct usb_bus {
	char *bus_name;                 /* stable id (PCI slot_name etc) */
};

/*
 * We do not support the full implementation of the linux usb device,
 * but only a partial one.
 * This one is similar to the struct usbd_device in FreeBSD
 * defined in usbdivar.h (pointed to by usbd_device_handle)
 */
struct usb_device {
#if 1	/* this is our simulated implementation */
        struct device dev;              /* Generic device interface */
	struct usb_device_descriptor descriptor;/* Descriptor */
					/* 'ddesc' on FreeBSD */

	int devnum;	/* addr. on USB bus, XXX only on 1 debug msg */
			/* uint8_t address on FreeBSD */

	/* XXX to be filled */
        char            devpath [16];   /* Use in messages: /port/port/... */
        char *product;                  /* iProduct string, if present */
        struct usb_bus *bus;            /* Bus we're part of */
        enum usb_device_speed   speed;  /* high/full/low (or error) */
#else	/* this is the linux implementation */
        int             devnum;         /* Address on USB bus */
			/* uint8_t address on FreeBSD */

        char            devpath [16];   /* Use in messages: /port/port/... */
			/* not on FreeBSD */

        enum usb_device_state   state;  /* configured, not attached, etc */
			/* config ? on FreeBSD */

        enum usb_device_speed   speed;  /* high/full/low (or error) */
			/* speed on FreeBSD */

        struct usb_tt   *tt;            /* low/full speed dev, highspeed hub */
        int             ttport;         /* device port on that tt hub */

        unsigned int toggle[2];         /* one bit for each endpoint
                                         * ([0] = IN, [1] = OUT) */

        struct usb_device *parent;      /* our hub, unless we're the root */
        struct usb_bus *bus;            /* Bus we're part of */
			/* bus on FreeBSD */

        struct usb_host_endpoint ep0;
			/* def_ep on FreeNSD */

        struct device dev;              /* Generic device interface */

        struct usb_device_descriptor descriptor;/* Descriptor */
			/* ddesc on FreeBSD */

        struct usb_host_config *config; /* All of the configs */

        struct usb_host_config *actconfig;/* the active configuration */
        struct usb_host_endpoint *ep_in[16];
        struct usb_host_endpoint *ep_out[16];

        char **rawdescriptors;          /* Raw descriptors for each config */

        unsigned short bus_mA;          /* Current available from the bus */
        u8 portnum;                     /* Parent port number (origin 1) */

        int have_langid;                /* whether string_langid is valid */
        int string_langid;              /* language ID for strings */

        /* static strings from the device */
        char *product;                  /* iProduct string, if present */
        char *manufacturer;             /* iManufacturer string, if present */
        char *serial;                   /* iSerialNumber string, if present */

        struct list_head filelist;
        struct class_device *class_dev;
        struct dentry *usbfs_dentry;    /* usbfs dentry entry for the device */

        /*
         * Child devices - these can be either new devices
         * (if this is a hub device), or different instances
         * of this same device.
         *
         * Each instance needs its own set of data structures.
         */

        int maxchild;                   /* Number of ports if hub */
        struct usb_device *children[USB_MAXCHILDREN];
#endif
};

/*
 * URB support, for asynchronous request completions
 */

struct urb;

struct pt_regs;	// XXX maybe somewhere else

/*
 * The handler called at the end of a transfer.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
typedef void (*usb_complete_t)(struct urb *);
#else
typedef void (*usb_complete_t)(struct urb *, struct pt_regs *);
#endif

/*
 * urb->transfer_flags:
 */
#define URB_SHORT_NOT_OK        0x0001  /* report short reads as errors */
#define URB_ISO_ASAP            0x0002  /* iso-only, urb->start_frame
                                         * ignored */
#define URB_NO_TRANSFER_DMA_MAP 0x0004  /* urb->transfer_dma valid on submit */
#define URB_NO_SETUP_DMA_MAP    0x0008  /* urb->setup_dma valid on submit */
#define URB_NO_FSBR             0x0020  /* UHCI-specific */
#define URB_ZERO_PACKET         0x0040  /* Finish bulk OUT with short packet */
#define URB_NO_INTERRUPT        0x0080  /* HINT: no non-error interrupt
                                         * needed */

struct usb_iso_packet_descriptor {
        unsigned int offset;
        unsigned int length;            /* expected length */
        unsigned int actual_length;
        unsigned int status;
};

struct urb {
	/*
	 * Fields for FreeBSD emulation:
	 *    +	the array of sizes is separate so we need to allocate
	 *	it explicitly (at interrupt time, this means NOWAIT).
	 *    + memory is associated to the xfer, so we _might_ need
	 *	to copy if someone scrambles us.
	 *    + there is a usbd_private_handle which is the argument to
	 *	the callback, but in fact this is the 'context' argument.
	 * note, we don't use the 'pipe' here but rely on the usbd_pipe_handle
	 * recorded in the struct linux_usb_descriptor.
	 * This should be fixed linking the urb to the pipes instead.
	 */
	int	n_frames;	/* how many frames */
	uint16_t *frlengths;	/* the array of sizes */
#if 0
        /* private: usb core and host controller only fields in the urb */
        struct kref kref;               /* reference count of the URB */
        spinlock_t lock;                /* lock for the URB */
        void *hcpriv;                   /* private data for host controller */
        int bandwidth;                  /* bandwidth for INT/ISO request */
        atomic_t use_count;             /* concurrent submissions counter */
        u8 reject;                      /* submissions will fail */

        /* public: documented fields in the urb that can be used by drivers */
        struct list_head urb_list;      /* list head for use by the urb's
                                         * current owner */
#endif
        struct usb_device *dev;         /* (in) pointer to associated device */
        unsigned int pipe;              /* (in) pipe information */
        int status;                     /* (return) non-ISO status */
        unsigned int transfer_flags;    /* (in) URB_SHORT_NOT_OK | ...*/
        char *transfer_buffer;          /* (in) associated data buffer */
        dma_addr_t transfer_dma;        /* (in) dma addr for transfer_buffer */
        int transfer_buffer_length;     /* (in) data buffer length */

	/* XXX must set this */
        int actual_length;              /* (return) actual transfer length */
#if 0
        unsigned char *setup_packet;    /* (in) setup packet (control only) */
        dma_addr_t setup_dma;           /* (in) dma addr for setup_packet */
#endif
        int start_frame;                /* (modify) start frame (ISO) XX unused*/
        int number_of_packets;          /* (in) number of ISO packets */
        int interval;                   /* (modify) transfer interval
                                         * (INT/ISO) */
#if 0
        int error_count;                /* (return) number of ISO errors */
#endif
        void *context;                  /* (in) context for completion */
        usb_complete_t complete;        /* (in) completion routine */
        struct usb_iso_packet_descriptor iso_frame_desc[0];
                                        /* (in) ISO ONLY */
};

/**
 * usb_fill_bulk_urb - macro to help initialize a bulk urb
 * @urb: pointer to the urb to initialize.
 * @dev: pointer to the struct usb_device for this urb.
 * @pipe: the endpoint pipe
 * @transfer_buffer: pointer to the transfer buffer
 * @buffer_length: length of the transfer buffer
 * @complete_fn: pointer to the usb_complete_t function
 * @context: what to set the urb context to.
 *
 * Initializes a bulk urb with the proper information needed to submit it
 * to a device.
 */
static inline void usb_fill_bulk_urb (struct urb *urb,
                                      struct usb_device *dev,
                                      unsigned int pipe,
                                      void *transfer_buffer,
                                      int buffer_length,
                                      usb_complete_t complete_fn,  
                                      void *context)
{
	// spin_lock_init(&urb->lock);
        urb->dev = dev;
        urb->pipe = pipe;
        urb->transfer_buffer = transfer_buffer;
        urb->transfer_buffer_length = buffer_length;
        urb->complete = complete_fn;
        urb->context = context;
}

/**
 * usb_fill_int_urb - macro to help initialize a interrupt urb
 * @urb: pointer to the urb to initialize.
 * @dev: pointer to the struct usb_device for this urb.
 * @pipe: the endpoint pipe
 * @transfer_buffer: pointer to the transfer buffer
 * @buffer_length: length of the transfer buffer 
 * @complete_fn: pointer to the usb_complete_t function
 * @context: what to set the urb context to.
 * @interval: what to set the urb interval to, encoded like
 *      the endpoint descriptor's bInterval value.
 * 
 * Initializes a interrupt urb with the proper information needed to submit
 * it to a device.
 * Note that high speed interrupt endpoints use a logarithmic encoding of
 * the endpoint interval, and express polling intervals in microframes
 * (eight per millisecond) rather than in frames (one per millisecond).
 */
static inline void usb_fill_int_urb (struct urb *urb,
                                     struct usb_device *dev,
                                     unsigned int pipe,
                                     void *transfer_buffer,
                                     int buffer_length,
                                     usb_complete_t complete_fn,
                                     void *context,
                                     int interval)
{
	// spin_lock_init(&urb->lock);
        urb->dev = dev;
        urb->pipe = pipe;
        urb->transfer_buffer = transfer_buffer;
        urb->transfer_buffer_length = buffer_length;
        urb->complete = complete_fn;
        urb->context = context;
        if (dev->speed == USB_SPEED_HIGH)
                urb->interval = 1 << (interval - 1);
        else
                urb->interval = interval;
        urb->start_frame = -1;
}

/*
 * Descriptor for the current status of an interface.
 * On FreeBSD, probably a partially matching structure is
 * struct usbd_interface
 */
struct usb_interface {
        /* array of alternate settings for this interface,
         * stored in no particular order */
        struct usb_host_interface *altsetting;
        struct usb_host_interface *cur_altsetting;      /* the currently
                                         * active alternate setting */
        unsigned num_altsetting;        /* number of alternate settings */

#if 0
        int minor;                      /* minor number this interface is
                                         * bound to */
        enum usb_interface_condition condition;         /* state of binding */
#endif
        struct device dev;              /* interface specific device info */
        struct class_device *class_dev;
};

struct usb_device *usb_get_dev(struct usb_device *dev);
void usb_put_dev(struct usb_device *dev);

int usb_clear_halt(struct usb_device *dev, int pipe);
int usb_set_interface(struct usb_device *dev, int ifnum, int alternate);

void *usb_buffer_alloc (struct usb_device *dev, size_t size,
        gfp_t mem_flags, dma_addr_t *dma);

void usb_buffer_free (struct usb_device *dev, size_t size,
        void *addr, dma_addr_t dma);

/// maybe init_urb is exposed
// void usb_init_urb(struct urb *urb);
struct urb *usb_alloc_urb(int iso_packets, gfp_t mem_flags);
void usb_free_urb(struct urb *urb);
int usb_submit_urb(struct urb *urb, gfp_t mem_flags);
int usb_unlink_urb(struct urb *urb);
void usb_kill_urb(struct urb *urb);

static inline void *usb_get_intfdata (struct usb_interface *intf)
{
        return dev_get_drvdata (&intf->dev);
}

static inline void usb_set_intfdata (struct usb_interface *intf, void *data)
{
        dev_set_drvdata(&intf->dev, data);
}

struct usb_interface *usb_get_intf(struct usb_interface *intf);

void usb_put_intf(struct usb_interface *intf);


int usb_string(struct usb_device *dev, int index,
        char *buf, size_t size);

/* devfs support... */
/**
 * usb_make_path - returns stable device path in the usb tree
 * @dev: the device whose path is being constructed
 * @buf: where to put the string
 * @size: how big is "buf"?
 */
int usb_make_path (struct usb_device *dev, char *buf, size_t size);

#endif /* _LINUX_USB_H */
